#include "incs.h"
#include "class_tree.h"
#include "tree.h"
#include "network.h"
#include "utils.h"
#include "functions.h"
#include "node_nt.h"
#include "node_term.h"
#include "net_utils.h"

Class_Tree::~Class_Tree()
{
  
}

Class_Tree::Class_Tree(int n_batch, FILE* fp, int dbg) 
  : Tree(n_batch, fp, dbg) 
{
  strcpy(class_str, "Class_Tree");
  gather_nodes();
  n_conn = get_n_params();
  printf("n_params = %d\n", n_conn);
}


Class_Tree::Class_Tree(int n_batch, 
				   int num_expert, 
				   int num_depth,
				   int num_ip, 
				   int num_op,
				   int dbg,
				   int xprt_optim,
				   int gate_optim,
				   real _tol,
				   int _itmax,
				   int _msteps,
				   int _rich,
				   int _seed,
				   int _gate_ensemble,
				   int _xprt_ensemble,
				   int _gate_hyper_epochs,
				   int _xprt_hyper_epochs
				   )
  : Tree(n_batch, num_expert, num_depth,
	 num_ip,
	 num_op,
	 dbg,
	 _tol,
	 _itmax,
	 _msteps,
	 _rich,
	 _seed),	 
	gate_ensemble(_gate_ensemble),
	xprt_ensemble(_xprt_ensemble),
	gate_hyper_epochs(_gate_hyper_epochs),
	xprt_hyper_epochs(_xprt_hyper_epochs)
{
  if(n_expert > 1) {
	node = build_recursive_tree((Node*) NULL, 
								n_expert,
								&n_levels,
								size_batch,
								xprt_optim,
								gate_optim);
  }
  else
	node = build_1layer_tree((Node*) NULL, 
							 size_batch,
							 xprt_optim,
							 gate_optim);
  strcpy(class_str, "Class_Tree");
  gather_nodes();
  n_conn = get_n_params();
  printf("n_params = %d\n", n_conn);

}

Node* Class_Tree::build_recursive_tree(Node* parent,
									 int num_children, 
									 int* current_depth, 
									 int num_train, 
									 int xprt_optim,
									 int gate_optim)
{
  if(debug > 0)
    printf("building gate %d\n", *current_depth);
  Network* gate;
  gate = get_new_class_net(gate_ensemble,
						   gate_optim,
						   num_train,
						   n_input, 
						   num_children, 
						   optim_iterations,
						   msteps,
						   optim_richness,
						   optim_tolerance,
						   gate_hyper_epochs,
						   debug);
  Node** children = new Node* [num_children];

  if(debug > 0)
    printf("building class_tree %d\n", *current_depth);
  Node_NT* this_node = new Node_NT(num_train, 
								   num_children, 
								   n_input,
								   n_output,
								   gate, 
								   children, 
								   debug,
								   get_new_id());
  for(int xp = 0; xp < num_children; xp ++){
    if(*current_depth == 1){
      Network* expert;
	  expert = get_new_class_net(xprt_ensemble,
								 xprt_optim,
								 num_train,
								 n_input,
								 n_output,
								 optim_iterations,
								 msteps,
								 optim_richness,
								 optim_tolerance,
								 xprt_hyper_epochs,
								 debug);
      children[xp] = new Node_Term(num_train, 
								   n_input,
								   n_output,
								   debug,
								   get_new_id(),
								   expert);
	  
      if(debug > 0)
		printf("building expert %d\n", *current_depth);
    }
    else
      {
		*current_depth -=1;
		children[xp] = build_recursive_tree(this_node,
											num_children, 
											current_depth, 
											num_train, 
											xprt_optim,
											gate_optim);
		*current_depth += 1;
	  }
  }
  for(int xp1 = 0 ; xp1 < num_children; xp1 ++)
    children[xp1]->set_parent(this_node);
  this_node->set_parent(parent);
  return this_node;
}

Node* Class_Tree::build_1layer_tree(Node* parent,
								  int num_train, 
								  int xprt_optim,
								  int gate_optim)
{
  Network* expert;
  expert = get_new_class_net(xprt_ensemble,
							 xprt_optim,
							 num_train,
							 n_input,
							 n_output,
							 optim_iterations,
							 msteps,
							 optim_richness,
							 optim_tolerance,
							 xprt_hyper_epochs,
							 debug);
  
  Node_Term* this_node = new Node_Term(num_train, 
									   n_input,
									   n_output,
									   debug,
									   get_new_id(),
									   expert);
  return this_node;
}

Class_Tree::Class_Tree(Class_Tree* old)
  : Tree((Tree*) old)
{
  strcpy(class_str, "Class_Tree");

  node = Tree::build_recursive_tree((Node_NT*)old->node, (Node*)NULL);
  gather_nodes();
  n_conn = get_n_params();
  printf("n_params = %d\n", n_conn);
}

Class_Tree::Class_Tree(const Class_Tree& old)
  : Tree(old)

{
  strcpy(class_str, "Class_Tree");

  node = Tree::build_recursive_tree((Node_NT*)old.node, (Node*)NULL);
  gather_nodes();
  n_conn = get_n_params();
  printf("n_params = %d\n", n_conn);
}


void Class_Tree::write_outputs(real** inputs, real** out, 
							 real** targets,
							 real** mom2, FILE* fp_res, int ntrain) {
  go_forward_batch(ntrain, inputs, out, mom2);
  fprintf(fp_res,"\n inputs = [\n");
  for(int n1 = 0; n1 < ntrain; n1 ++) {
    for(int i =0; i < n_input; i ++){
      fprintf(fp_res,"%f ", inputs[n1][i]);
    } 
    fprintf(fp_res,"\n");
  }
  fprintf(fp_res,"];\n");
  fprintf(fp_res,"\n outputs = [\n");
  for(int n = 0; n < ntrain; n ++) {
    for(int i1 =0; i1 < n_output; i1 ++){
      fprintf(fp_res,"%f ", out[n][i1]);
    }
    for(int i2 =0; i2 < n_output; i2 ++){
      fprintf(fp_res,"%f ", targets[n][i2]);
    }
	for(int i =0; i < n_output; i ++){
      fprintf(fp_res,"%f ", mom2[n][i]);
    }
    fprintf(fp_res,"\n");
  }
  fprintf(fp_res,"];\n");

}

double Class_Tree::compute_like(int N, real** out, real** targ) {
  double lk = 0.0;
  for(int n = 0; n < N; n ++) {
	for(int o = 0; o < n_output; o ++) {
	  lk -= targ[n][o] * Slog(out[n][o]);
	}
  }
  return lk;
}

double Class_Tree::compute_error(int N, real** in, real** targ) {
  go_forward_batch(N, in, outputs);
  double err = 0.0;
  /*  for(int n = 0; n < N; n ++) {
	for(int o = 0; o < n_output; o ++) {
	  double arg = 0.0;
	  if(targ[n][o] > 0.0)
		arg = -targ[n][o] * Slog(outputs[n][o]);
	  err += arg;
	}
  }
  err /= N;*/
  for(int n = 0; n < N; n ++) {
	int best = -1;
	real max = -FLT_MAX;
	for(int o = 0; o < n_output; o ++) {
	  if(outputs[n][o] > max) { 
		max = outputs[n][o]; 
		best = o;
	  }
	}
	for(int o = 0; o < n_output; o ++) {
	  if(targ[n][o] > 0.0) {
		if(best == o) err += 0.0;
		else err += 1.0;
	  }
	}
  }
  err /= N;
  return err;
}
