#include "incs.h"
#include "data.h"
#include "utils.h"

Data::Data (FILE* data_file, int num_points, 
			int num_inputs, int num_targets ):
  data_fp(data_file),
  n_points(num_points),
  n_inputs(num_inputs+1),
  n_targets(num_targets)
{
  input_matrix = RealMatrix2(n_points, n_inputs);
  target_matrix = RealMatrix2(n_points, n_targets);

  float tmp;
  for(int point = 0; point < n_points; point ++){
    input_matrix[point][0] = 1.0;
    for(int ipdim = 1; ipdim < n_inputs; ipdim ++){
      if(feof(data_fp))
	error(1, "End of data file !\n");
      fscanf(data_fp,"%f",&tmp);
      input_matrix[point][ipdim] = (real) tmp;
    }
    for(int opdim = 0; opdim < n_targets; opdim ++){
      if(feof(data_fp))
	error(1, "End of data file !\n");
      fscanf(data_fp,"%f",&tmp);
      target_matrix[point][opdim] = (real) tmp;
    }
  }
}

Data::~Data()
{
  
}

real Data::get_next_input ( int pointer, real* input )
{
  return input[pointer];
}

real Data::get_next_target ( int pointer, real* target )
{
  return target[pointer];
}

void Data::print ( )
{
  for(int point = 0; point < n_points; point ++){
    printf("inputs = ");
    for(int ipdim = 0; ipdim < n_inputs; ipdim ++){
      printf("%f ", input_matrix[point][ipdim]);
    }
    printf("\n");
    printf("targets = ");
    for(int opdim = 0; opdim < n_targets; opdim ++){
      printf("%f ", target_matrix[point][opdim]);
    }
    printf("\n");
  }
}


  
  
real** Data::get_inputs(int offset)
{
  return input_matrix + (offset*n_inputs);
}

real** Data::get_targets(int offset)
{
  return target_matrix + (offset*n_targets);
}

  
