/**************************************************************************/
/**************************************************************************/
/**************************************************************************/
/***									***/ 
/*** 		Authors : Dan Kershaw & Gary Cook			***/
/*** 		Cambridge University Engineering Department		***/
/*** 		Trumpington Street, Cambridge, CB2 1PZ, England.	***/
/***									***/
/*** 		getArgs.h : header for command line processing          ***/
/***									***/
/***		=======   COPYRIGHT NOTICE   =======			***/
/***		Copyright (c) 1996, Dan Kerhaw & Gary Cook		***/
/*** 		All rights reserved.					***/
/***									***/
/**************************************************************************/
/**************************************************************************/
/**************************************************************************/

#ifndef _getArgs_h_

#define _getArgs_h_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include "incs.h"


// typedef a structure rather than class 
// used to make things self contained
typedef struct _optInfo {
  
  struct _optInfo *next;      
  char *flag;           // option flag
  char *defValue;       // default settings
  char *help;           // help message
  char *errMsg;         // error message found by command line processing
  
} OptInfo ;

class getArgs {

private:
  
  // The private object variables

  int *argc;               // standard number of args (argc mirror)
  char **argv;             // standard arguments      (argv mirror)
  bool success;            // flag to see if arguments read in succesfully
  OptInfo *optInfo;        // stores the option information

  // The private object functions

  // update the arguments
  void updateArgs( int rm_cnt );

  // check the arguments for errors
  bool checkArgError( char *argument,
		      char *flag, 
		      int defValue, 
		      char *helpMsg ) ;

  bool checkArgError( char *argument,
		      char *flag, 
		      float defValue, 
		      char *helpMsg ) ;

  bool checkArgError( char *argument,
		      char *flag, 
		      char *defValue, 
		      char *helpMsg ) ;

  bool checkArgError( char *flag, 
		      bool defValue, 
		      char *helpMsg ) ;

  // add information to the optInfo linked list 
  void addOptionInfo( char *flag, 
		      char *defValue, 
		      char *help,
		      char *errMsg = NULL );

  // print error messages from the optInfo linked list
  void printErrors(void);

  // print help message using the optInfo linked list
  void printHelp(void);

public:

  // constructor
  getArgs(int *numArgs, char **args) {
    
    argc = numArgs;
    argv = args;
    success = true;
    optInfo = new OptInfo[1];
    optInfo->next = NULL;

  };

  // command line processing of flagged argments

  // process integer arguments
  int intArg( char *flag, 
	      char *helpMsg,
	      int defValue = 0 );

  int intArg( char *flag, 
	      int defValue = 0 ); 

  // process single precision arguments
  float floatArg( char *flag, 
		  char *helpMsg,
		  float defValue = 0.0 ); 

  float floatArg( char *flag, 
		  float defValue = 0.0 ); 

  // process character string arguments
  char *stringArg( char *flag, 
		   char *helpMsg,
		   char *defValue = "None"); 

  char *stringArg( char *flag, 
		   char *defValue = "None"); 
  
  // process flag arguments
  bool switchArg( char *flag, 
		char *helpMsg,
		bool defValue = false );

  bool switchArg( char *flag, 
		bool defValue = false );

  // print the usage
  void printUsage( char *firstline = NULL );

  // do a final check of success and errors
  void finalCheck( int numReqArgs = 0, char *firstline = NULL );

} ;

#endif
