#include <time.h>
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "incs.h"
#include "getArgs.h"
#include "data.h"
#include "utils.h"
#include "class_tree.h"
#include "net_utils.h"
#include "functions.h"

extern "C" int gethostname(char*, ...);

#define DEF_MSTEPS 10
#define DEF_TOL 0.001
#define DEF_ITMAX 10
#define DEF_PRIOR_VAR 1.0
#define DEF_LAMBDA 2
#define FALSE 0
#define TRUE 1
#define DEF_NINP 1
#define DEF_N_OUTPUT 2
#define DEF_DEPTH 1
#define DEF_SEED -1
#define DEF_MAX_EPOCHS 200
#define DEF_GATE_RAND 1.0
#define DEF_XPRT_RAND 1.0
#define DEF_NTRAIN 100
#define DEF_NCV 00
#define DEF_DEBUG 0
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 1
#endif
int main(int argc, char **argv) 
{
  char* progName = argv[0];  
  char* version = "0.93";
  getArgs comline = getArgs( &argc, argv );
  // process the command line
  bool more_help = (bool) FALSE;
  if(argc == 1){
    more_help = (bool) TRUE;
  }
  bool help = comline.switchArg( "-help",
				 "\tDisplay command line usage",
				 false );
  int seed = comline.intArg( "-seed",
			     "\trandom seed\t\t",
			     DEF_SEED );
  if(seed < -1)
    error(1, "seed = %d : seed must be greater than 0", seed);

  int n_input = comline.intArg( "-n_input",
				"input vector dimension\t",
				DEF_NINP );
  if(n_input < 1)
    error(1, "n_input = %d : Number of inputs must be > 0", n_input);
  
  int n_output = comline.intArg( "-n_output",
				 "target vector dimension\t",
				 DEF_N_OUTPUT );

  int npoints = comline.intArg( "-n_train",
			       "number of training points",
			       DEF_NTRAIN );
  char* out_file = comline.stringArg( "-out_file",
									  "Weights out file stem\t",
									  "-");
  char* train_file = comline.stringArg( "-train_file",
					"Training data file\t",
					"-");
  // show command line usage
  if ( help || more_help) {
    comline.printUsage( "test [options]" );
    exit( EXIT_SUCCESS );
  }
  
  comline.finalCheck( 0, "test [options]" );

  // defaults

  int itmax = 3;
  int msteps = DEF_MSTEPS;
  int rich = 0;
  real tol = DEF_TOL;
  int xprt_optim = 1;
  int gate_optim = 1;
  int debug = 0;
  int n_epochs = DEF_MAX_EPOCHS;
  int depth = 1;
  int n_in = n_input + 1;
  int n_experts = 1;
  printf("using %d experts\n", n_experts);

  real lambda = DEF_LAMBDA;
  real prior_var = DEF_PRIOR_VAR;
  
  if(strcmp(train_file, "-")==0)
    {
      error(0,"You must specify a training data file");
      comline.printUsage( "test [options]" );      
    }
  if(strcmp(out_file, "-")==0)
    {
      error(0,"You must specify an output weights file\n");
      comline.printUsage( "test [options]" );
    }
  char hostname[1024];
  time_t now;
  time(&now);
  assert(gethostname(hostname, 1024)==0);
  printf("Program %s version %s started on %s\n at %s \n",
		 progName, version,
		 hostname,ctime(&now));
  
  FILE* data_fp;

  data_fp = openFile(train_file, "r");

  Data* train_data_ptr;
  train_data_ptr = new Data(data_fp, npoints, n_input, n_output);
  Data& train_data = *train_data_ptr;

  real** orig_inputs = train_data.get_inputs(0);
  real** orig_targets = train_data.get_targets(0);
  time_t start_time;
  time_t stop_time;
  start_time = time(NULL);
  if(seed == -1) seed = get_seed();
  set_seed(seed);

	Class_Tree* class_tree;
	
	class_tree = (Class_Tree*) new Class_Tree(npoints, 
										n_experts, 
										depth, 
										n_input, 
										n_output,
										debug,
										xprt_optim,
										gate_optim,
										tol,
										itmax,
										msteps,
										rich,
										seed,
										0,0,1,1);
	
	real** inputs = orig_inputs;
	real** targets = orig_targets;

	int epochs = 0;


	FILE* out_fp = openFile(out_file, "w");
	class_tree->rand_posts(npoints, inputs, targets);
	epochs = class_tree->train(n_epochs, 0,
							 inputs, targets,
							 out_fp, npoints, 1);
	printf("\n saving params to file %s\n", out_file);
	class_tree->write_weights(out_fp);
	fclose(out_fp);
	delete class_tree;
	stop_time = time(NULL);
	double time_elapsed = stop_time - start_time;
	printf("time elapsed %f s\n ",
		   time_elapsed);
  time(&now);
  printf("Program %s ended at %s \n",progName,ctime(&now));
}


