#include "incs.h"
#include "net_regress_ens.h"
#include "lu.h"
#include "utils.h"
#include "functions.h"
#include "net.h"
NET_Regress_Ens::NET_Regress_Ens(int n_batch, 
								 int num_ip, 
								 int num_op,
								 real _lambda,
								 real _prior_var,
								 int dbg,
								 int _msteps,
								 real _tol
								 )
  : n_weights(num_ip * num_op),
    msteps(_msteps),
    tol(_tol),
    Net(n_batch, num_ip, num_op, dbg),
    lambda(_lambda),
    prior_var(_prior_var)
{
  weights = new double[n_output * n_input];
  var_acc = new double[n_output];
  variance = new double[n_output];
  for(int i = 0; i < n_output; i ++) {
    variance[i] = prior_var/lambda; // set to some value initially
  }
  class_str = new char[1024];
  hessian = new float**[n_output];
  for(int i = 0; i < n_output; i ++) {
    hessian[i] = new float*[n_input];
    for(int o = 0; o < n_input; o ++) {
      hessian[i][o] = new float[n_input];
    }
  }
  inv_hess = new float**[n_output];
  for(int i = 0; i < n_output; i ++) {
    inv_hess[i] = new float*[n_input];
    for(int o = 0; o < n_input; o ++) {
      inv_hess[i][o] = new float[n_input];
    }
  }
  alpha = new double[n_output];
  for(int o = 0; o < n_output; o ++) {
    alpha[o] = 0.0;
  }
  ssw = new double[n_output];
  hess_trace = new float[n_output];
  uncertainty = new double[n_batch];
  for(int n = 0; n < n_batch; n ++) {
	uncertainty[n] = 0.0;
  }
  strcpy(class_str, "NET_Regress");
}
NET_Regress_Ens::NET_Regress_Ens(const NET_Regress_Ens& old)
  : n_weights(old.n_weights),
    msteps(old.msteps),
    tol(old.tol),
    Net(old),
    lambda(old.lambda),
    prior_var(old.prior_var),
	hyper_epochs(old.hyper_epochs)
{
  weights = new double[n_output * n_input];
  for(int n = 0; n < n_weights; n ++) {
	weights[n] = old.weights[n];
  }
  var_acc = new double[n_output];
  variance = new double[n_output];
  for(int i = 0; i < n_output; i ++) {
    variance[i] = old.variance[i];
  }
  class_str = new char[1024];
  hessian = new float**[n_output];
  for(int i = 0; i < n_output; i ++) {
    hessian[i] = new float*[n_input];
    for(int o = 0; o < n_input; o ++) {
      hessian[i][o] = new float[n_input];
    }
  }
  inv_hess = new float**[n_output];
  for(int i = 0; i < n_output; i ++) {
    inv_hess[i] = new float*[n_input];
    for(int o = 0; o < n_input; o ++) {
      inv_hess[i][o] = new float[n_input];
    }
  }
  for(int o = 0; o < n_output; o ++) {
	for(int i = 0; i < n_output; i ++) {
	  for(int i2 = 0; i2 < n_output; i2 ++) {
		hessian[o][i][i2] = old.hessian[o][i][i2];
		inv_hess[o][i][i2] = old.inv_hess[o][i][i2];
	  }
	}
  }
  alpha = new double[n_output];
  for(int o = 0; o < n_output; o ++) {
    alpha[o] = old.alpha[o];
  }
  ssw = new double[n_output];
  for(int o = 0; o < n_output; o ++) {
	ssw[o] = old.ssw[o];
  }  
  hess_trace = new float[n_output];
  for(int o = 0; o < n_output; o ++) {
	hess_trace[o] = old.hess_trace[o];
  }  
  uncertainty = new double[size_batch];
  for(int n = 0; n < size_batch; n ++) {
	uncertainty[n] = old.uncertainty[n];
  }
  strcpy(class_str, "NET_Regress");
}

NET_Regress_Ens::NET_Regress_Ens(int n_batch, FILE* fp) 
: Net(n_batch, fp)
{
  strcpy(class_str, "NET_Regress");
  weights = new double[n_output * n_input];
  var_acc = new double[n_output];
  variance = new double[n_output];
  hessian = new float**[n_output];
  for(int i = 0; i < n_output; i ++) {
    hessian[i] = new float*[n_input];
    for(int o = 0; o < n_input; o ++) {
      hessian[i][o] = new float[n_input];
    }
  }
  inv_hess = new float**[n_output];
  for(int i = 0; i < n_output; i ++) {
    inv_hess[i] = new float*[n_input];
    for(int o = 0; o < n_input; o ++) {
      inv_hess[i][o] = new float[n_input];
    }
  }
  alpha = new double[n_output];
  for(int o = 0; o < n_output; o ++) {
    alpha[o] = 0.0;
  }
  hess_trace = new float[n_output];
   uncertainty = new double[n_batch];
   for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
      double tmp;
      fscanf(fp, "%lf ", &tmp);
      weights[i*n_input +j] = tmp;
    }
  }  
  for(int j = 0; j < n_output; j ++){
    double tmp;
    fscanf(fp, "%lf ", &tmp);
    variance[j] = tmp;
  }
  for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
	  for(int k = 0; k < n_input; k ++){
		float tmp;
		fscanf(fp, "%f ", &tmp);
		inv_hess[i][j][k] = tmp;
	  }
	}
  }
}  
  
NET_Regress_Ens::~NET_Regress_Ens()
{
}

void NET_Regress_Ens::randomize_weights(float offset)
{
  for(int wt_num = 0; wt_num < n_weights; wt_num ++) {
    weights[wt_num] = offset * 2.0* (get_drand() - 0.5);
  }
  for(int o = 0; o < n_output; o ++) {
    alpha[o] = offset*offset;
  }
}

void NET_Regress_Ens::randomize_weights2(float offset)
{
  for(int wt_num = 0; wt_num < n_weights; wt_num ++) {
	weights[wt_num] =  (1.0 + offset * 2.0*(get_drand() - 0.5)) * weights[wt_num]  ;
  }
}

void NET_Regress_Ens::go_forward(float* in, float* out, int point)
{
  for(int op = 0;  op < n_output;  op ++) {
    out[op] = 0.0;
  }
  for(int op = 0;  op < n_output;  op ++) {
    for(int ip = 0; ip < n_input; ip ++) {
      out[op] += weights[op*n_input + ip] * in[ip];
    }
  }
  for(int op = 0;  op < n_output;  op ++) {
    outputs[point][op] = out[op];
  }
}

void NET_Regress_Ens::go_forward_log(float* in, float* out, int point)
{
  for(int op = 0;  op < n_output;  op ++) {
    out[op] = 0.0;
  }
  for(int op = 0;  op < n_output;  op ++) {
    for(int ip = 0; ip < n_input; ip ++) {
      out[op] += weights[op*n_input + ip] * in[ip];
    }
  }
  for(int op = 0;  op < n_output;  op ++) {
    outputs[point][op] = out[op];
  }
}
void NET_Regress_Ens::go_forward(float* in, float* out, 
								 float* m2, int point)
{
  for(int op = 0;  op < n_output;  op ++) {
	out[op] = 0.0;
  }
  for(int op = 0;  op < n_output;  op ++) {
	for(int ip = 0; ip < n_input; ip ++) {
	  out[op] += weights[op*n_input + ip] * in[ip];
	}
  }
  for(int op = 0; op < n_output; op ++) {
	double unc = 0.0;
	for(int i = 0; i < n_input; i ++) {
	  for(int i2 = 0; i2 < n_input; i2 ++) {
		unc += in[i] * inv_hess[op][i][i2] * in[i2];
	  }
	}
	m2[op] = unc + variance[op] + (out[op] * out[op]);
  }
  for(int op = 0;  op < n_output;  op ++) {
	 outputs[point][op] = out[op];
	 mom2[point][op] = m2[op];
   }
}

void NET_Regress_Ens::get_gradient(double* grad) {
  for(int o = 0; o < n_output; o ++) {
	for(int i = 0; i < n_input; i ++) {
      int n = o * n_input + i;
	  grad[n] = alpha[o] * weights[n];
	}
  }
  for(int point = 0; point < batch_size; point ++) {
    go_forward(inputs[point], outputs[point], point);
    int wt_num = 0;
    for(int op = 0; op < n_output; op ++) {
      double err = (outputs[point][op] - targets[point][op]) / variance[op] ;
      for(int ip =0; ip < n_input; ip ++) {
		grad[wt_num] +=  jpost[point] * err * inputs[point][ip];
		wt_num ++;
      }
    }
  }
}

void NET_Regress_Ens::get_gradient(double* grad, double* fn) {
  *fn = 0.0;
 for(int o = 0; o < n_output; o ++) {
    for(int i = 0; i < n_input; i ++) {
      int n = o * n_input + i;
      grad[n] = alpha[o] * weights[n];
    }
  }
 for(int o = 0; o < n_output; o ++) {
   ssw[o] = 0.0;
 }
 for(int o = 0; o < n_output; o ++) {
   for(int i = 0; i < n_input; i ++) {
	 int n = o * n_input + i;
	 ssw[o] += weights[n];
   }
 }
  for(int o = 0; o < n_output; o ++) {
	(*fn) += alpha[o]*ssw[o];
  }

  for(int point = 0; point < batch_size; point ++) {
    go_forward(inputs[point], outputs[point], point);
    int wt_num = 0;
    for(int op = 0; op < n_output; op ++) {
      double err = (outputs[point][op] - targets[point][op]) /variance[op] ;
      for(int ip =0; ip < n_input; ip ++) {
		grad[wt_num] +=  jpost[point] * err * inputs[point][ip];
		wt_num ++;
      }
      (*fn) += jpost[point] * err * err;
    }
  }
}

double NET_Regress_Ens::get_function() {
  double fn = 0.0;
 for(int o = 0; o < n_output; o ++) {
   ssw[o] = 0.0;
 }
 for(int o = 0; o < n_output; o ++) {
   for(int i = 0; i < n_input; i ++) {
	 int n = o * n_input + i;
	 ssw[o] += weights[n];
   }
 }
  for(int o = 0; o < n_output; o ++) {
	(fn) += alpha[o]*ssw[o];
  }
  for(int point = 0; point < batch_size; point ++) {
    go_forward(inputs[point], outputs[point], point);
    for(int op = 0; op < n_output; op ++) {
      double err = outputs[point][op] - targets[point][op] ;
      fn += jpost[point] * err * err;
    }
  }
  return fn;
}

void NET_Regress_Ens::read_weights(FILE* fp) {
  int ninp;
  int noutp;
  fscanf(fp, "%s\n", class_str);
  fscanf(fp, "%d %d\n", &noutp, &ninp);
  for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
      float tmp;
      fscanf(fp, "%f ", &tmp);
      weights[i*n_input +j] = tmp;
    }
  }
  for(int j = 0; j < n_output; j ++){
    float tmp;
    fscanf(fp, "%f ", &tmp);
    variance[j] = tmp;
  }
}
void NET_Regress_Ens::write_weights(FILE* fp)
{
  fprintf(fp, "%s\n", class_str);
  fprintf(fp, "%d %d\n", n_output, n_input);
  for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
      fprintf(fp, "%f ", weights[i*n_input +j]);
    }
    fprintf(fp, "\n");
  }
  for(int j = 0; j < n_output; j ++){
    fprintf(fp, "%f ", variance[j]);
  }
}

void NET_Regress_Ens::compute_gen_probs(int point, float* target)
{
  // each output is modelled by a Gaussian
  // with mean (out[i]-targ[i]) and variance[i]
  float arg = 0.0;
  for(int i = 0; i < n_output; i ++) {
    float error1 = outputs[point][i] - target[i];
    arg += error1 * error1 / variance[i];
	arg += Slog(variance[i]);
  }
  arg += uncertainty[point];
  gen_prob[point] = -arg;

  for(int op = 0; op < n_output; op ++) 
	gen_prob[point] -= alpha[op] * ssw[op];
}

void NET_Regress_Ens::compute_hessian(int n_points, 
									  float** in, 
									  float* wts,
									  float*** hess)
{
  for(int o = 0; o < n_output; o ++) {
    for(int i = 0; i < n_input; i ++) {
      for(int i2 = 0; i2 < n_input; i2 ++) {
		hess[o][i][i2] = 0.0;
      }
    }
	for(int i = 0; i < n_input; i ++) {
	  hess[o][i][i] = alpha[o];
	}
  }
  
  for(int n = 0; n < n_points; n ++) {
    for(int o = 0; o < n_output; o ++) {
      float w = wts[n] / variance[o];
      for(int i = 0; i < n_input; i ++) {
		for(int i2 = 0; i2 < n_input; i2 ++) {
		  hess[o][i][i2] += w  * in[n][i] * in[n][i2];
		}
      }
    }
  }
}

double NET_Regress_Ens::train(int n_points, float** in, 
							  float** targ)
{
  for(int n = 0; n < n_points; n ++) {
    jpost[n] = 1.0;
  }
  // this is a hack to get round the / by variance in
  // train_params...
  for(int o = 0; o < n_output; o ++) {
	variance[o] = 1.0;
  }
  double cost = train_params(n_points, in, targ);
  go_forward_batch(n_points, in, outputs);
  update_variance(n_points, outputs, targ, jpost);
  return cost;
}
void NET_Regress_Ens::init(int n_points, 
						   float** in, 
						   float** targ,
						   float* wts)
{
  for(int n = 0; n < n_points; n ++) {
    jpost[n] = Sexp(wts[n]);
  }
  go_forward_batch(n_points, in, outputs);
  update_variance(n_points, outputs, targ, jpost);
  double cost =  train_params(n_points, in, targ, jpost);
  if(debug > 0)
	printf("cost = %f\n", cost);
}

double NET_Regress_Ens::train(int n_points, 
							  float** in, 
							  float** targ,
							  float* wts)
{
  targets = targ;
  inputs = in;
  for(int n = 0; n < n_points; n ++) {
    jpost[n] = Sexp(wts[n]);
  }
  double cost =  train_params(n_points, in, targ, jpost);
  return cost;
}

void NET_Regress_Ens::update_hyperparameters(int n_points, 
											 float** in, 
											 float** targ,
											 float* wts) {
  go_forward_batch(n_points, in, outputs);
  for(int o = 0; o < n_points; o ++) {
	jpost[o] = Sexp(wts[o]);
  }

  compute_hessian(n_points, in, jpost, hessian);
  for(int o = 0; o < n_output; o ++) {
	if(isposdef(hessian[o], n_input) == 0) {
	  error(1, "hessian is not positive definite");
	}
  }
  for(int o = 0; o < n_output; o ++) {
	double det = 0.0;
	det = lu_inverse(hessian[o], inv_hess[o], n_input);
	hess_trace[o] = 0.0;
	for(int i = 0; i < n_input; i ++) {
	  hess_trace[o] += inv_hess[o][i][i];
	}
  }
  for(int n = 0; n < n_points; n ++) {
	uncertainty[n] = 0.0;
  } 
  for(int n = 0; n < n_points; n ++) {
	 for(int o = 0; o < n_output; o ++) {
	   for(int i = 0; i < n_input; i ++) {
		 for(int i2 = 0; i2 < n_input; i2 ++) {
		   uncertainty[n] += in[n][i] * inv_hess[o][i][i2] * in[n][i2];
		 }
	   }
	 }
  }
  for(int o = 0; o < n_output; o ++) {
	 ssw[o] = 0.0;
	 for(int i = 0; i < n_input; i ++) {
	   ssw[o] += weights[o*n_input + i] * weights[o*n_input + i];
	 }
  }
  double alpha_0 = 0.1;
  double ssw_0 = 0.1;
  for(int o = 0; o < n_output; o ++) {
	 alpha[o] = (n_input + alpha_0) / (ssw[o] + ssw_0 + hess_trace[o]);
	 //	 printf("xprt alpha = %f\n", alpha[o]);
   }
  update_variance(n_points, outputs, targ, jpost);
}

void NET_Regress_Ens::update_variance(int n_points, 
									  float** out, 
									  float** targ,
									  float* wts) {
  for(int i = 0; i < n_output; i ++) {
    var_acc[i] = 0.0;
  }
  double sum_wts = 0.0;
  for(int point2 = 0; point2 < n_points; point2 ++){
    sum_wts += wts[point2];
    for(int op =0; op < n_output; op ++) {
      float v = (targ[point2][op] - out[point2][op]);
      float v2 = v * v;
	  v2 += uncertainty[point2];
      var_acc[op] += wts[point2] * v2;
    }
  }
  for(int i = 0; i < n_output; i ++) {
    variance[i] = (var_acc[i] + prior_var) / (sum_wts + lambda);
  }
  if(debug > 0){
    for(int i = 0; i < n_output; i ++) {
      printf("variance = %f %f \n", variance[i], sum_wts);
    }
  }
}

double NET_Regress_Ens::get_joint(int point, float* target)
{
  // each output is modelled by a Gaussian
  // with mean (out[i]-targ[i]) and variance[i]
  float arg = 0.0;
  for(int i = 0; i < n_output; i ++) {
    float error1 = outputs[point][i] - target[i];
    arg += error1 * error1 / variance[i];
	arg += Slog(variance[i]);
  }
  arg += uncertainty[point];
  double joint = -arg;

  for(int op = 0; op < n_output; op ++) 
	joint -= alpha[op] * ssw[op];
  return joint;
}
void NET_Regress_Ens::update_variance(int n_points, 
								  float** out, 
								  float** targ) {
  for(int i = 0; i < n_output; i ++) {
    var_acc[i] = 0.0;
  }
  for(int point2 = 0; point2 < n_points; point2 ++){
    for(int op =0; op < n_output; op ++) {
      float v = (targ[point2][op] - out[point2][op]);
      var_acc[op] += v * v;
    }
  }
  for(int i = 0; i < n_output; i ++) {
	variance[i] = (var_acc[i]) / n_points;
  }
}

