#ifndef NODE_H
#define NODE_H

class Node
{
public:
  Node(int n_batch, 
	   int num_ip, 
	   int num_op,
	   int dbg,
	   int node_id);
  Node(int n_batch, FILE* fp); 
  Node(const Node& old_node);
  virtual void reset_nodes() = 0;
  int num_input() {return n_input;};
  int num_output() {return n_output;};
  void set_parent(Node* par) { parent = par;};
  Node* get_parent() { return parent; };  
  real get_cond_post(int n) { return cond_post[n];};
  void set_cond_post(int n, real p) { cond_post[n] = p;};
  void set_joint_post(int n, real p) { joint_post[n] = p;};
  real get_joint_post(int n) { return joint_post[n];};
  void set_cond_prob(int n, real p) { cond_prob[n] = p;}
  real get_cond_prob(int n) { return cond_prob[n];};
  char* get_class() {return class_str;};
  int get_id() {return id;};
  void set_id(int new_id) { id = new_id; };
  real* get_jpost_vec() { return joint_post;};
  // virtual functions after this line
  // ---------------------------------
 
  virtual ~Node() ;
  virtual void update_params(int N, real** in, real** targ) =0;
  virtual void update_hypers(int N, real** in, real** targ) = 0;
  virtual void train(int N, real** in, real** targ) = 0;
  virtual void init(int N, real** in, real** targ) = 0;
  virtual int is_terminal() = 0;
  virtual void randomize_weights(real offset) = 0;
  virtual void randomize_weights(real xp_off,real gt_off) = 0;
  virtual void write_weights(FILE* fp) = 0;
  virtual void read_weights(FILE* fp) = 0;
  virtual void print_posteriors(int n) = 0;
  virtual real get_joint_prob(int n) = 0;
  virtual void set_joint_prob(real p, real* in, int n) = 0;
  virtual real set_dens(int point, real* in, real* targ) = 0;
  virtual real get_dens() = 0;
  virtual void set_posts(int point) = 0;
protected:
  Node* parent;
  char* class_str;
  int visited;
  int n_input;
  int n_output;
  int id;
  int size_batch;
  int debug;
  real* cond_post;
  real* joint_post;
  real* joint_prob;
  real* cond_prob;
  real prob_dens;
};

#endif


