#include "incs.h"
#include "node_term.h"
#include "utils.h"
#include "functions.h"
#include "net_utils.h"
#include "node_nt.h"

Node_Term::~Node_Term()
{
}

Node_Term::Node_Term(Node& old) :
  Node(old)
{
  // takes a non terminal and merges into a terminal
  strcpy(class_str, "Node_Term");
  Node_NT* old_nt = (Node_NT*) &old;
  int n_child = old_nt->get_num_children();
  Node* old_child = old_nt->get_child(0);
  if(strcmp(old_child->get_class(), "Node_Term") == 0) {
	Node_Term* old_term = (Node_Term*) old_child;
	Network* old_expert = old_term->get_expert();
	expert = get_new_net(old_expert);
  }
  else {
	error(1, "Wrong kind of node in Node_Term(Node& old) %s\n", old_child->get_class());
  }
}

Node_Term::Node_Term(Node& old, Network* default_expert) :
  Node(old)
{
  // takes a non terminal and merges into a terminal
  strcpy(class_str, "Node_Term");
  expert = get_new_net(default_expert);
}

Node_Term::Node_Term(int n_batch, 
					 int num_ip, 
					 int num_op,
					 int dbg,
					 int node_id,
					 Network* _expert)
  : Node(n_batch, num_ip, num_op, dbg, node_id),  
   expert(_expert)
{
  strcpy(class_str, "Node_Term");
}

Node_Term::Node_Term(const Node_Term& old)
  : Node(old)
{
  strcpy(class_str, "Node_Term");
  expert = get_new_net(old.expert);
}

Node_Term::Node_Term(int n_batch, FILE* fp) : 
  Node(n_batch, fp)

{
  strcpy(class_str, "Node_Term");

  fscanf(fp, "%d\n", &id);
  fscanf(fp, "%d %d\n", &n_input, &n_output);
  char* tmp = CharVector(1024);
  fscanf(fp, "%s\n", tmp);
  if(strcmp(tmp, "Expert")!=0)
    error(1, "I was expecting an expert here...\n");
  
  fscanf(fp, "%s\n", tmp);
  expert = get_new_net(size_batch, fp, tmp);
  if(expert == NULL) {
	error(1, "Expert is NULL !\n");
  }
}

void Node_Term::randomize_weights(real xprt_offset, real gate_offset)
{
  expert->randomize_weights(xprt_offset);
}

void Node_Term::randomize_weights(real offset)
{
  expert->randomize_weights(offset);
}

void Node_Term::set_joint_prob(real prob, real* in, int n) {
  joint_prob[n] = prob;
}

// ---------------------------------------------------------
// various training routines
// ---------------------------------------------------------

void Node_Term::init(int N, real** in, real** targ)
{
  expert->init(N, in, targ, joint_post);
}
// ---------------------------------------------------------
void Node_Term::update_hypers(int N, real** in, 
							  real** targ)
{
  if(test_prune(joint_post, N) == 1) {
	//	printf("pruning out\n");
	return;
  }
  expert->update_hyperparameters(N, in, targ, joint_post);
}
void Node_Term::update_params(int N, real** in, 
							  real** targ)
{
  if(test_prune(joint_post, N) == 1) {
	//	printf("pruning out\n");
	return;
  }
  expert->train(N, in, targ, joint_post);
}

void Node_Term::train(int N, real** in, real** targ)
{
  if(test_prune(joint_post, N) == 1) {
	//	printf("pruning out\n");
	return;
  }
  update_params(N, in, targ);
  update_hypers(N, in, targ);
}
// ---------------------------------------------------------
// things to do with computing posteriors...
// ---------------------------------------------------------
// ---------------------------------------------------------
// miscellaneous i/o stuff 
// ---------------------------------------------------------

void Node_Term::write_weights(FILE* fp)
{
  fprintf(fp, "%s\n", class_str);
  fprintf(fp, "%d\n", id);
  fprintf(fp, "%d %d\n", n_input, n_output);
  fprintf(fp, "Expert\n");
  expert->write_weights(fp);
}

void Node_Term::read_weights(FILE* fp)
{
  fscanf(fp, "%s\n", class_str);
  fscanf(fp, "%d\n", &id);
  fscanf(fp, "%d %d\n", &n_input, &n_output);
  char* tmp = CharVector(1024);
  fscanf(fp, "%s\n", tmp);
  if(strcmp(tmp, "Expert")!=0)
    error(1, "I was expecting an expert here...\n");
  expert->read_weights(fp);
  FreeCharVector(&tmp);
}

void Node_Term::print_posteriors(int n)
{
  printf("%s: point %d cp=%f jp=%f cpr=%f jpr=%f\n",
		 class_str,
		 n, Sexp(cond_post[n]), Sexp(joint_post[n]),
		 Sexp(cond_prob[n]), Sexp(joint_prob[n]));
}

void Node_Term::reset_nodes() {
  visited = 0;
}

real Node_Term::get_dens() {
  return prob_dens;
}

real Node_Term::set_dens(int point, real* in, real* targ) {
  real dens = 0.0;
  dens = expert->get_joint(in, targ);
  prob_dens = dens;
  return dens;
}

void Node_Term::set_posts(int point) {
  if(parent == NULL) {
	joint_post[point] = 0.0;
	cond_post[point] = 0.0;
  }
  return;
}
