#include "tree.h"

void Tree::randomize_weights(real xprt_offset, 
							 real gate_offset) 
{
  node_xp_offset = xprt_offset;
  node_gt_offset = gate_offset;
  node->randomize_weights(xprt_offset, gate_offset);
}


double* get_var(int n_points, real** in, int n_input) {
  double* vacc = new double[n_input];
  double* macc = new double[n_input];
  for(int i = 0; i < n_input; i ++) {
	macc[i] = 0.0;
	vacc[i] = 0.0;
  }
  for(int n = 0; n < n_points; n++) {
	for(int j = 0; j < n_input; j ++) {
	  vacc[j] += in[n][j]*in[n][j];
	  macc[j] += in[n][j];
	}
  }
  double* ivar = new double[n_input];
  for(int j = 0; j < n_input; j ++) {
	macc[j] /= n_points;
	ivar[j] = sqrt(vacc[j]/n_points - macc[j]* macc[j]);
  }
  return ivar;
}

void Tree::rand_terms(int n_points, real** in, real** targ) {
  double* ivar = get_var(n_points, in, n_input);
  double* ovar = get_var(n_points, targ, n_output);

  for(int m = 0; m < n_terminals; m ++) {
	Net* term = (Net*) terminals[m];
	double* term_params = term->get_weights();
	for(int o = 0; o < n_output; o ++) {
	  for(int i = 0 ; i < n_input; i ++) {
		int w = o * n_input + i;
		double scale = 0.0;
		if(ivar[i] > 0.0) 
		  scale = ovar[o] / ivar[i];
		term_params[w] = scale * gasdev(&seed);
	  }
	}
  }
}

void Tree::rand_gates() {
  for(int g = 0; g < n_gates; g ++) {
	Net* gate = (Net*) gates[g];
	double* g_params = gate->get_weights();
	int n_out = gate->num_output();
	for(int o = 0; o < n_out; o ++) {
	  for(int i = 0 ; i < n_input; i ++) {
		int w = o * n_input + i;
		double scale = 1.0 / sqrt(n_input);
		g_params[w] = scale * gasdev(&seed);
	  }
	}
  }
}
void Tree::wta_init(int ntrain, real** inputs, real** targets) {
  int n_epochs = 100;
  int init_epoch = 0;
  double log_like = 0.0;
  real double_diff = 2.0;
  real old_diff = 0.0;
  real average_diff = 2.0;
  int average_length = 3;
  double diff = 0.0;
  real av = 0.0;
  int rec_epoch = init_epoch;
  int epoch = 0;
  while(((fabs(average_diff) > 1e-5) && (epoch < n_epochs))) {
    if(epoch > 1) {
      if((epoch % average_length) == 0){
		average_diff = av / average_length;
		double_diff = fabs(old_diff - average_diff);
		old_diff = average_diff;
		av = 0.0;
      }
      av += diff;
    }
	get_hard_posts(ntrain, inputs, targets);
	//	node->update_params(ntrain, inputs, targets);
	node->train(ntrain, inputs, targets);

	go_forward_batch(ntrain, inputs, outputs);
    double old_log_like = log_like;
	log_like = compute_like(ntrain, outputs, targets);
	double error = compute_error(ntrain, inputs, targets);
	
    diff = fabs(log_like - old_log_like);

    printf("%4d\t%4.3g\t%4.3g\t%4.3g\t%4.3g\n",
		   rec_epoch,
		   error,
		   log_like, 
		   diff,
		   average_diff);
    epoch ++;
	rec_epoch ++;
  }

  
}
void Tree::rand_posts(int n_points, real** in, real** targ) {

  real* joint_post = RealVector(n_terminals);

  for(int n = 0; n < n_points; n ++) {
	real sum = 0.0;
	for(int xp = 0 ; xp < n_terminals; xp ++) {
	  joint_post[xp] = get_drand();
	  sum += joint_post[xp];
	}
	for(int xp2 = 0 ; xp2 < n_terminals; xp2 ++) {
	  joint_post[xp2] /= sum;
	  check_arg(Slog(joint_post[xp2]), "jp");
	  term_nodes[xp2]->set_joint_post(n, Slog(joint_post[xp2]));
	}
	propagate_posts(term_nodes, n_terminals, n);

  }
  for(int xp = 0 ; xp < n_terminals; xp ++) {
	term_nodes[xp]->init(n_points, in, targ);
  }
  if(strcmp(node->get_class(), "Node_NT") == 0)
	 ((Node_NT*)node)->train_gates(n_points, in);
  FreeRealVector(&joint_post);
}
