#include "tree.h"

// ------------------------------------------------------------

// things to do with global training

// ------------------------------------------------------------

double Tree::mac_train(int n_points, real** in, real** targ)
{
  //  printf("training params\n");
  mac_inputs = in;
  mac_targets = targ;
  batch_size = n_points;
  if(mac_pass == 0) {
	int np = 0;
	count_params(node, &np);
	n_weights = np;
	weights = RealVector(np);
	mac_weights = RealVector(np+1);
	old_weights = RealVector(np);
	grad = RealVector(np);
	mac_pass ++;
	itmax = 3;
	tol = 0.001;
  }
  double cost = 0.0;
  batch_size = n_points;
  get_weights(weights);
  for(int n1 = 1; n1 <= a_n; n1 ++) {
    mac_weights[n1] = weights[n1-1];
  }

  macoptII(mac_weights, a_n, itmax, tol);

  for(int n = 1; n <= a_n; n ++) {
    weights[n-1] = mac_weights[n];
  }
  set_weights(weights);
  return cost;
}
void Tree::get_weights(real* weights) {
  int wt_num = 0;
  for(int t = 0; t < n_terminals; t ++) {
	real* wt = terminals[t]->get_weights();
	int n_w = terminals[t]->get_num_weights();
	for(int w = 0; w < n_w; w ++) {
	  if(wt_num > n_weights)
		error(1, "wt_num too big %d %d", wt_num, n_weights);
	  weights[wt_num] = wt[w];
	  wt_num ++;
	}
  }
  for(int nt = 0; nt < n_gates; nt ++) {
	real* wt2 = gates[nt]->get_weights();
	int n_w2 = gates[nt]->get_num_weights();
	for(int w2 = 0; w2 < n_w2; w2 ++) {
	  if(wt_num > n_weights)
		error(1, "wt_num too big");
	  weights[wt_num] = wt2[w2];
	  wt_num ++;
	}
  }
}

void Tree::set_weights(real* weights) {
  int wt_num = 0;
  for(int t = 0; t < n_terminals; t ++) {
	real* wt = terminals[t]->get_weights();
	int n_w = terminals[t]->get_num_weights();
	for(int w = 0; w < n_w; w ++) {
	  if(wt_num > n_weights)
		error(1, "wt_num too big");
	  wt[w] = weights[wt_num];
	  wt_num ++;
	}
  }
  for(int nt = 0; nt < n_gates; nt ++) {
	real* wt2 = gates[nt]->get_weights();
	int n_w2 = gates[nt]->get_num_weights();
	for(int w2 = 0; w2 < n_w2; w2 ++) {
	  if(wt_num > n_weights)
		error(1, "wt_num too big");
	  wt2[w2] = weights[wt_num];
	  wt_num ++;
	}
  }
}
void Tree::get_gradient(double* g) {
  int wt_num = 0;
  for(int t = 0; t < n_terminals; t ++) {
	real* wt = terminals[t]->get_grad(batch_size, mac_inputs,
									  mac_targets);
	int n_w = terminals[t]->get_num_weights();
	for(int w = 0; w < n_w; w ++) {
	  if(wt_num > n_weights)
		error(1, "wt_num too big");
	  g[wt_num] = wt[w];
	  wt_num ++;
	}
  }
  for(int nt = 0; nt < n_gates; nt ++) {
	Node_NT* nd = (Node_NT*) nt_nodes[nt];
	nd->get_new_gate_targets(batch_size);
	real** gt = nd->get_gate_targets();
	real* wt2 = gates[nt]->get_grad(batch_size, mac_inputs,
									gt);
	int n_w2 = gates[nt]->get_num_weights();
	for(int w2 = 0; w2 < n_w2; w2 ++) {
	  if(wt_num > n_weights)
		error(1, "wt_num too big");
	  g[wt_num] = wt2[w2];
	  wt_num ++;
	}
  }

}

double Tree::get_function() {
  get_posts(batch_size, mac_inputs, mac_targets);
  return like;
}

void Tree::dfunc(double* new_weights, double* dw) {
  get_weights(weights);
  for(int w = 0; w < n_weights; w ++) {
	old_weights[w] = weights[w];
  }
  
  for(int k = 1; k <= a_n; k ++) {
    weights[k-1] = new_weights[k];
  }
  set_weights(weights);

  get_gradient(grad);

  for(int j = 1; j <= a_n; j ++) {
    dw[j] = grad[j-1];
  }
  for(int w = 0; w < n_weights; w ++) {
	weights[w] = old_weights[w];
  }
  set_weights(weights);
}

double Tree::func(double* _weights) {
  double fn = 0.0;
  get_weights(weights);
  for(int w = 0; w < n_weights; w ++) {
	old_weights[w] = weights[w];
  }
  for(int j = 1; j <= a_n; j ++) {
    weights[j-1] = _weights[j];
  }
  set_weights(weights);
  fn = get_function();
  for(int w = 0; w < n_weights; w ++) {
	weights[w] = old_weights[w];
  }
  set_weights(weights);
  return fn;
}

