#include "tree.h"

void Tree::get_sa_posts(int n_points, real** in, real** targ)
{
  real over_sum = LZERO;
  real lik = 0.0;
  real ax = 0.0;
  real jp_sum = LZERO;
  for(int n = 0; n < n_points; n ++) {
	node->set_joint_prob(0.0, in[n], n);
	real sum = LZERO;
	real lik_n = LZERO;
	for(int ch = 0; ch < n_terminals; ch ++) {
	  real dens = terminals[ch]->get_joint(in[n], targ[n]);
	  real j = term_nodes[ch]->get_joint_prob(n);
	  jp_sum = LAdd(jp_sum, j);
	  real jp = j + dens;
	  jp *= beta; // this is the SA bit
	  lik_n = LAdd(lik_n, jp);
	  term_nodes[ch]->set_joint_post(n, jp);
	  if(jp > LSMALL)
		sum = LAdd(sum, jp);
	}
	lik += lik_n;
	for(int ch2 = 0; ch2 < n_terminals; ch2 ++) {	
	  real jp = term_nodes[ch2]->get_joint_post(n) - sum;
	  term_nodes[ch2]->set_joint_post(n, jp);
	  real dens = terminals[ch2]->get_joint(in[n], targ[n]);
	  real j = term_nodes[ch2]->get_joint_prob(n);
	  ax += Sexp(jp) * (j + dens);
	  over_sum = LAdd(over_sum, jp);
	}
	propagate_posts(term_nodes, n_terminals, n);
  }
  aux = ax;
  like = lik;
}
int Tree::sa_train(int n_epochs, 
				   int init_epoch,
				   real** inputs, 
				   real** targets,
				   FILE* fp_out, 
				   int ntrain) {
  int epoch = 0;
  double log_like = 0.0;
  real double_diff = 2.0;
  real old_diff = 0.0;
  real average_diff = 2.0;
  int average_length = 3;
  double diff = 0.0;
  real av = 0.0;
  int rec_epoch = init_epoch;
  real inc_beta = 0.01; // means that beta will be 1.0 aftr 20
  // epochs
  beta = 0.5;
  //  beta = 1.0;
  while((beta < 1.0) ||((fabs(average_diff) > 1e-4) && (epoch < n_epochs))) {
    if(epoch > 1) {
      if((epoch % average_length) == 0){
		average_diff = av / average_length;
		double_diff = fabs(old_diff - average_diff);
		old_diff = average_diff;
		av = 0.0;
      }
      av += diff;
    }
    double old_log_like = like;
	if(beta >= 1.0)
	  beta = 1.0;
	else
	  beta += inc_beta;

	train(1000, 0, inputs, targets, fp_out, ntrain, 1);	

	log_like = like;
	double error = compute_error(ntrain, inputs, targets);
    diff = fabs(log_like - old_log_like);
    printf("sa train: %4d\t%f\t%4.3g\t%4.3g\t%4.3g\t%4.3g\t\n",
		   rec_epoch,
		   beta,
		   error,
		   log_like, 
		   diff,
		   average_diff);

	fprintf(fp_out,
			"%4d\t%4.3g\t%4.3g\t%4.3g\t%4.3g\n", 
			rec_epoch,
			error,
			log_like, 
			diff,
			average_diff
			);
    epoch ++;
	rec_epoch ++;
  }
  return 1;
}

int Tree::cv_sa_train(int n_epochs, 
					  int init_epoch,
					  real** trn_inputs, 
					  real** trn_targets,
					  real** tst_inputs, 
					  real** tst_targets,
					  FILE* fp_out, 
					  int ntrain,
					  int ntest) {
  int epoch = 0;
  double log_like = 0.0;
  real double_diff = 2.0;
  real old_diff = 0.0;
  real average_diff = 2.0;
  int average_length = 3;
  double diff = 0.0;
  real av = 0.0;
  real inc_beta = 0.01; // means that beta will be 1.0 aftr 20
  // epochs
  beta = 0.5;
  //  beta = 1.0;
  double min_err = FLT_MAX;
  while((beta < 1.0) ||((fabs(average_diff) > 1e-5) && 
						(epoch < n_epochs))) {
    if(epoch > 1) {
      if((epoch % average_length) == 0){
		average_diff = av / average_length;
		double_diff = fabs(old_diff - average_diff);
		old_diff = average_diff;
		av = 0.0;
      }
      av += diff;
    }
    double old_log_like = like;
	if(beta >= 1.0)
	  beta = 1.0;
	else
	  beta += inc_beta;
	double last_min_err = min_err;
	min_err = cv_train2(1000, 0, trn_inputs, trn_targets, 
						tst_inputs,
						tst_targets, fp_out, ntrain, ntest,
						last_min_err);	
	
	log_like = like;
	double trn_err = 
	  compute_error(ntrain, trn_inputs, trn_targets);
	double tst_err = 
	  compute_error(ntest, tst_inputs, tst_targets);

    diff = fabs(log_like - old_log_like);
    printf("sa train: %4d\t%f\t%4.3g\t%4.3g\t%4.3g\t%4.3g\n",
		   epoch,
		   beta,
		   min_err,
		   trn_err,
		   tst_err,
		   log_like);
    epoch ++;
  }
  return 1;
}


double Tree::cv_train2(int n_epochs, 
				   int init_epoch,
				   real** trn_inputs, 
				   real** trn_targets,
				   real** tst_inputs, 
				   real** tst_targets,
				   FILE* fp_out, 
				   int ntrain,
				   int ntest,
				   double min_err) {
  double diff = 0.1;
  int rec_epoch = init_epoch;
  int epoch = 0;
  double err = compute_error(ntest, tst_inputs, tst_targets);
  int verb = 0;
  if(verb > 0) 
	printf("\n%4d\t%4.3g\t%4.3g\t%4.3g\t", 
		   rec_epoch,
		   min_err,
		   err,
		   diff); 
  int best_epoch = 1;

  while((epoch < 60) ||((epoch < best_epoch * 1.5) 
						&& (epoch < n_epochs))) {
    epoch ++;
	rec_epoch ++;
	if(err < min_err) {
	  // save current set of weights...
	  min_err = err;
	  rewind(fp_out);
	  write_weights(fp_out);
	  if(verb > 0)
		printf("%f (s)\n", min_err);
	  best_epoch = epoch;
	}
	else{
	  if(verb > 0)
		printf("\n");
	}
	// train on training set
	double old_like = like;
	get_sa_posts(ntrain, trn_inputs, trn_targets);
	double like_diff = like - old_like;
	node->train(ntrain, trn_inputs, trn_targets);
	if((like_diff < 0.01)) {
	  get_sa_posts(ntrain, trn_inputs, trn_targets);
	  mac_train(ntrain, trn_inputs, trn_targets);
	}
	// get error on training set
	//	double trn_err = compute_error(ntrain, trn_inputs, trn_targets);
	// get error on test set
	double old_error = err;
	err = compute_error(ntest, tst_inputs, tst_targets);
    diff = old_error - err;
    // output results
	if(verb > 0) 
	  printf("%4d\t%4.3g\t%4.3g\t%4.3g\t%d\t", 
			 rec_epoch,
			 min_err,
			 err,
			 diff,
			 best_epoch);
  }
  printf("\n");
  return min_err;
}
