
#ifndef _utils_h_

#define _utils_h_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "incs.h"

// open a file, with error checking
FILE *openFile(const char *filename,
	       const char *options,
	       char *errMsg = "Error:");

// close a file
void closeFile(FILE *stream,
	       char *errMsg = "Error:");

// do a binary read with error checking
int bfread( void *ptr, 
	       int size, 
	       int nobj, 
	       FILE *stream,
	       char *errMsg = "Error:");

// do a binary write with error checking
void bfwrite( const void *ptr,
	      int size, 
	      int nobj, 
	      FILE *stream,
	      char *errMsg = "Error:");

// report errors
void error(int errcode, char *message, ...);

/* float vectors and matrices */
void*   Smalloc(int);
void	Sfree(void*);
void    FreeFloatMatrix2(float ***);
void    FreeFloatVector(float **);
float   *FloatVector(int);
void    FreeFloatMatrix3(float ****);
float   **FloatMatrix2(int,int);
float   ***FloatMatrix3(int,int,int);
float   **floatmatrix2(int,int);
float   ***floatmatrix3(int,int,int);
float   *floatvector(int);



/* ints */

int     *intvector(int);
int     *IntVector(int);
void    FreeIntVector(int **);
int     **IntMatrix2(int,int);
void    FreeIntMatrix2(int ***);

/* chars */

char     *charvector(int);
char     *CharVector(int);
void    FreeCharVector(char **);
/* unsigned chars */

unsigned char     *ucharvector(int);
unsigned char     *UcharVector(int);
  


void    FreeDoubleMatrix2(double ***);
void    FreeDoubleVector(double **);
double   *DoubleVector(int);
void    FreeDoubleMatrix3(double ****);
double   **DoubleMatrix2(int,int);
double   ***DoubleMatrix3(int,int,int);
double   **doublematrix2(int,int);
double   ***doublematrix3(int,int,int);
double   *doublevector(int);


void    FreeRealMatrix2(real ***);
void    FreeRealVector(real **);
real   *RealVector(int);
void    FreeRealMatrix3(real ****);
real   **RealMatrix2(int,int);
real   ***RealMatrix3(int,int,int);
real   **realmatrix2(int,int);
real   ***realmatrix3(int,int,int);
real   *realvector(int);


#endif
